#!/usr/bin/env python

""" Prototype wrapper for C++ GIPS pipeline, version SHISv0

This script wrapper sources the appropriate FBF files, 
and spins off the C++ PipelineAdapter as a separate process
connected via pipes that masquerade as stdin/stdout on the C++ end.
(non meta-)data is passed along via memory maps.

Author: Maciek Smuga-Otto <maciek@ssec.wisc.edu>

version $Id: GipsHarness.py,v 1.11 2005/07/08 20:21:48 maciek Exp $

Copyright UW/SSEC, ALL RIGHTS RESERVED, 2004
Space Science and Engineering Center
University of Wisconsin - Madison, USA

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
"""

from FBF import FBF
from Numeric import *
from os import *
from glob import glob
import struct

GPLOT_ON = 0 ### - set to 0 if Gnuplot is not installed, or not desired.
UUID_SIZE = 16 # bytes

def get_temp_location():
    return environ.get('TMP', '/tmp')

# the bridge callback architecture:
# 
# a list of concrete bridges (such as fbf2mmap) is handed to the harness, which uses the provided callbacks.
# the default implementations of the callbacks do... nothing.

class bridge:
    def build(self):
        """ Tasks needed to build bridge, run before child process is forked.

        Include things like setting environmental variables, building mmaps...
        """
        pass

    def initialize(self, child):
        """ Tasks needed to run after child proces is forked, but before regular processing begins.

        This is usually left alone, except when building pipelines and the like.
        """
        pass
    
    def tear_down(self):
        """ Tasks needed to tear down bridge, run at very end. """
        pass
    
    def before_call(self, rec_num):
        """ Tasks to perform before each pipeline call, such as reading from fbf input files.
 
        Also, return (key,value) pair to include in message sent to executable via its stdin,
        such as record number to process, observation type, or the like
        """
        return None

    def after_call(self, rec_num):
        """ Tasks to perform after each pipeline call, such as writing to fbf output files. 
        """
        pass                          

    def num_records(self):
        """ Return max number of input records available, or max capacity of output if relevant.

        Otherwise (for non record based bridges), do not override the default of None.
        """
        return None                   

class fbf2mmap(bridge):
    def __init__(self, fbf, mmap, mmap_format):
        self.fbf_name = fbf
        self.mmap_name = mmap
        self.mmap_format = mmap_format

    def build(self):
        # open FBF file for reading
        full_fbf_name = glob('%s.*' % self.fbf_name)[0]
        self.fbf = FBF(full_fbf_name)
        self.fbf_fp = self.fbf.fp()
        ### old style:
        # create a memory map for a new FBF file to leverage the FBF library's
        # facility with reading/writing Numeric arrays as binary files.
        #self.mmap_fbf = FBF(stemname='GIPSTMP_'+ self.mmap_name, typename=self.mmap_format, grouping=[self.record_size,], dirname=get_temp_location())
        #self.mmap_fp = self.mmap_fbf.create()
        #self.mmap_fd = self.mmap_fp.fileno() # convenience method to get file descriptor

        # resize buffer to proper size.
        #z = zeros(self.record_size, {'complex16': Complex, 'complex8': Complex}.get(self.mmap_format, Float))
        #self.mmap_fbf.block_write(1, z)
        #self.mmap_fp.flush()
        #self.mmap_fp.seek(0)

        ### new style:
        # create raw file, to accommodate mixed data types (such as UUID at beginning)
        self.mmap_fp = file('GIPSTMP_'+self.mmap_name, 'w+b', 0)
        self.mmap_fd = self.mmap_fp.fileno()
        self.mmap_size = UUID_SIZE + self.fbf.record_size
        zerofill = '\x00' * self.mmap_size
        self.mmap_fp.write(zerofill)
        self.mmap_fp.flush()
        self.mmap_fp.seek(0)
        
        environ[self.mmap_name] = `self.mmap_fd`
        
    def tear_down(self):
        self.fbf.close()
        self.mmap_fp.unlink()
        self.mmap_fp.close()

    def num_records(self):
        return self.fbf.length()

    def before_call(self, rec_num):
        ### old style
        #data = self.fbf.read(rec_num)
        #self.mmap_fp.seek(0)
        # HACK!
        #self.mmap_fbf.block_write(1, data)
        #self.mmap_fbf.block_write(1, array(data, {'complex16': Complex, 'complex8': Complex}.get(self.mmap_format, Float)))

        ### new style
        self.fbf_fp.seek(self.fbf.record_size * (rec_num -1))
        data = self.fbf_fp.read(self.fbf.record_size)
        self.mmap_fp.seek(UUID_SIZE) # skip the UUID bytes
        self.mmap_fp.write(data)
        self.mmap_fp.flush()
	return None
        
class fbf2stdin(bridge):
    def __init__(self, fbf, keyname, typemap={}):
        self.fbf_name = fbf
        self.keyname = keyname
        self.typemap = typemap

    def build(self):
        full_fbf_name = glob('%s.*' % self.fbf_name)[0]
        self.fbf = FBF(full_fbf_name)

    def tear_down(self):
        self.fbf.close()

    def num_records(self):
        return self.fbf.length()

    def before_call(self, rec_num):
	raw = self.fbf.read(rec_num)
	return (keyname, typemap.get(raw, raw))

class mmap2fbf(bridge):
    def __init__(self, fbf, mmap, mmap_format, record_len):            
        self.fbf_name = fbf
        self.mmap_name = mmap
        self.record_len = record_len
        self.mmap_format = mmap_format
        if not mmap_format:
            self.mmap_format = fbf_format

    def build(self):
        # create FBF file for writing
        self.fbf = FBF("%s.%s.%s" % (self.fbf_name, self.mmap_format, self.record_len))
        self.fbf.create()
        self.fbf_fp = self.fbf.fp()

        ### old style
        # create a memory map for a new FBF file to leverage the FBF library's
        # facility with reading/writing Numeric arrays as binary files.
        #self.mmap_fbf = FBF(stemname='GIPSTMP_'+ self.mmap_name, typename=self.mmap_format, grouping=[self.record_size,], dirname=get_temp_location())
        #self.mmap_fp = self.mmap_fbf.create()
        #self.mmap_fd = self.mmap_fp.fileno() # convenience method to get file descriptor

        # resize buffer to proper size.
        #z = zeros(self.record_size, {'complex16': Complex, 'complex8': Complex}.get(self.mmap_format, Float))
        #self.mmap_fbf.block_write(1, z)
        #self.mmap_fp.flush()
        #self.mmap_fp.seek(0)

        ### new style
        self.mmap_fp = file('GIPSTMP_' + self.mmap_name, 'w+b', 0)
        self.mmap_fd = self.mmap_fp.fileno()
        self.mmap_size = UUID_SIZE + self.fbf.record_size
        zerofill = '\x00' * self.mmap_size
        self.mmap_fp.write(zerofill)
        self.mmap_fp.flush()
        self.mmap_fp.seek(0)

        environ[self.mmap_name] = `self.mmap_fd`

    def after_call(self, rec_num):
        #print return_msg #DEBUG
        #data = self.mmap_fbf.read(1)
        self.mmap_fp.seek(UUID_SIZE) # skip the UUID bytes
        data = self.mmap_fp.read(self.fbf.record_size)
        
	self.fbf_fp.seek(self.fbf.record_size * (rec_num - 1))
        self.fbf_fp.write(data)
        self.fbf_fp.flush()

    def tear_down(self):
        self.fbf.close()
        self.mmap_fp.unlink()
        self.mmap_fp.close()

class kvargs2env(bridge):
    def __init__(self, **kvargs):
        self.kvargs = kvargs

    def build(self):
        for key in self.kvargs:
            environ[key] = self.kvargs[key]

class staticStruct2mmap(bridge):
    def __init__(self, mmap, structlist):
        self.mmap_name = mmap
        self.structlist = structlist
        
    def build(self):
        self.mmap_fp = file('GIPSTMP_'+ self.mmap_name, 'w+b', 0)
        self.mmap_fd = self.mmap_fp.fileno()
        # print self.mmap_name + " is mapped to " + `self.mmap_fd` #DEBUG
        format_string = reduce(lambda x,y:x+y[0], self.structlist, '') # catenates the struct format chars into a single string
        values = [s[-1] for s in self.structlist]
        packed_struct = struct.pack(format_string, *values) # struct.pack accepts arguments in (format, val1, val2, ...) form
        self.mmap_fp.seek(UUID_SIZE) # skip UUID bytes
        self.mmap_fp.write(packed_struct)
        self.mmap_fp.flush()
        self.mmap_fp.seek(0)
        
        environ[self.mmap_name] = `self.mmap_fd`
        
    def tear_down(self):
        self.mmap_fp.unlink()
        self.mmap_fp.close()

class recNumRepeater(bridge):
    def before_call(self, rec_num):
        return rec_num
    
class aggregate2mmap(bridge):
    """ Aggregates one or more other (simplified) bridges to an mmap
    
     structlist consists of triples, the first element of which is a packing char, the second a name,
     and the third an instance of a bridge whose before_call method returns the appropriate value for a given record number"""
    def __init__(self, mmap, structlist):
        self.mmap_name = mmap
        self.structlist = structlist
        
    def build(self):
        for br in self.structlist:
            br[-1].build()
        
        self.mmap_fp = file('GIPSTMP_'+ self.mmap_name, 'w+b', 0)
        self.mmap_fd = self.mmap_fp.fileno()
        self.format_string = reduce(lambda x,y:x+y[0], self.structlist, '') # catenates the struct format chars into a single string
        
        environ[self.mmap_name] = `self.mmap_fd`
        
    def tear_down(self):
        for br in self.structlist:
            br[-1].tear_down()
            
        self.mmap_fp.unlink()
        self.mmap_fp.close()
    
    def before_call(self, rec_num):
        """ the last element of each triple is a class instance that defines __call__(self, rec_num)
        or, more simply, any function accepting a rec_num argument
        to echo rec_num, just supply lambda x: x
        """
        values = [s[-1].before_call(rec_num) for s in self.structlist]
        packed_struct = struct.pack(self.format_string, *values) # struct.pack accepts arguments in (format, val1, val2, ...) form

        self.mmap_fp.seek(UUID_SIZE) # skip UUID bytes
        self.mmap_fp.write(packed_struct)
        self.mmap_fp.flush()

        return None

    def num_records(self):
        raw_nrs = [s[-1].num_records() for s in self.structlist]
        no_null_nrs = [n for n in raw_nrs if n]
        if no_null_nrs:
            return min(no_null_nrs)
        else:
            return None

###### FIXME: Need to implement the multiple temperature FBF files to single MMAP bridge.
###### This may be done by means of a generic multi-FBF (single value record) to single MMAP bridge. 

class createPipeline(bridge):
    """ A specialized bridge that creates the pipeline.

    Run only once, after the child process is created but before regular processing.
    """
    
    def __init__(self, pipeline_name):
        self.name = pipeline_name

    def initialize(self, child):
        reply = child.converse("CREATE PIPELINE:" + self.name)
        
class harness:
    def __init__(self, executable, bridge_list):
        self.bridges = bridge_list

        for bridge in self.bridges:
            bridge.build()

        self.child = NewProcessCradle(executable)

        for bridge in self.bridges:
            bridge.initialize(self.child)
            
	self.num_records = min([b.num_records() for b in self.bridges if b.num_records()])

        # Determine max size of records
        self.max_record = -1
        
    def run(self, run_range):
        if run_range[1] == -1:
            run_range[1] = self.num_records

        # raw_input('press any key to cycle pipeline') #DEBUG
        
        for record in range(run_range[0], run_range[1]):
            send_msg_bits = [('TASK', record),]

            for bridge in self.bridges:
                send_msg_bits.append(bridge.before_call(record))            

	    send_msg = ' '.join(["%s:%s" % b for b in send_msg_bits if b])
            #print send_msg # DEBUG
            resp_msg = self.child.converse('PROCESS ' + send_msg)
            #print resp_msg # DEBUG
            
            for bridge in self.bridges:
                bridge.after_call(record)

    def __delete__(self):
        for bridge in self.bridges:
            bridge.tear_down()

############# Instrument-specific Utility objects

SHISobsTypeBridge = fbf2stdin('sceneMirrorPosition', 'obsType', 
    {
     63: 'INVALID',      # invalid entry
     65: 'AMBIENT',      # ambient blackbody
     69: 'OBS',          # earth observation
     72: 'HOT',          # hot blackbody
     83: 'SPACE'         # sky observation
    })

SHISscanDirectionBridge = fbf2stdin('scanDirection', 'PIPELINE',
    {
     66: 'BW',  # backwards
     70: 'FW'   # forwards
    })

BANDSIZE_MAP = { 
   2044: [('SHIS','LW'),],
   4088: [('SHIS','MW'), ('SHIS','SW')],
   1024: [('GIFTS','LW'),],
   2048: [('GIFTS','SMW'),]
}


################ Legacy: NewProcessCradle and friends.


def new_pipe_ends():
    """ create new pipe, return opened files on both sides """
    rfd, wfd = pipe()
    return fdopen(wfd, 'w'), fdopen(rfd, 'r')

class NewProcessCradle:
    """ Resource wrapper for spawned process that guarantees preservation of environment
    variables and filehandles.

    Initialization spawns a new process, binds pipes to allow for two-way communication 
    via other process's stdin & stdout. """
    
    def __init__(self, exe):
        print exe
        p2c_write, p2c_read = new_pipe_ends()   # this process -> child
        c2p_write, c2p_read = new_pipe_ends()   # child -> this process
        
        pid = fork()
        if (pid == 0):
            # child process
            p2c_write.close()
            c2p_read.close()

            dup2(p2c_read.fileno(), 0) # replace child process stdin with p2c_out
            dup2(c2p_write.fileno(), 1)  # and stdout with c2p_in
            
            execv(exe, [exe])
            
        else:
            # parent process (This python script)
            p2c_read.close()
            c2p_write.close()
            
            self.sender = p2c_write
            self.recver = c2p_read

    def sendMsg(self, msg):
        # print "sending: " + msg #DEBUG
        print >>self.sender, msg
        self.sender.flush()

    def converse(self, msg):
        self.sendMsg(msg)
        resp=self.recver.readline()
        #print "received: " + resp #DEBUG
        return resp

