/** 
*  \file Physics.h
*  \brief contains definitions in the Physics namespace
*  \author R.K.Garcia <rayg@ssec.wisc.edu>
*
*  \version $Id: Physics.h,v 1.8 2005/02/17 07:29:00 maciek Exp $
*
*  \par Copyright:
*  \verbatim
*
*  Copyright UW/SSEC, ALL RIGHTS RESERVED, 2004
*  Space Science and Engineering Center
*  University of Wisconsin - Madison, USA
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this library; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
*  \endverbatim
*/


#ifndef H_PHYSICS
#define H_PHYSICS

/**
*  \brief Physics namespace, including standard types and formulae for 
*  radiances, wavenumbers, temperatures, times, voltages.
*/
namespace gips 
{
    namespace physics 
    { 
	
	/**
	 * floating point time representation
	 * Units: seconds
	 */
	typedef double seconds_t;
	
	/**
	 * floating point temperature representation
	 * Units: Kelvin
	 */
	typedef double kelvin_t;
	
	/**
	 * floating point wavenumber representation
	 * Units: wavenumbers
	 */
	typedef double wnum_t;
	
	/** 
	 *inverse of wavenumbers - centimeters.
	 */
	typedef double cm_t;
	
	/** 
	 * Radiance_t value
	 * Units: 
	 */
	typedef double radiance_t;
	
	
	/** \fun planckRadiance
	 * \param T temperature in Kelvin
	 * \param wnum wavenumber
	 * \todo Implement me!
	 */
//	inline radiance_t planckRadiance( kelvin_t T, wnum_t wnum )
//	{
//	    return radiance_t( 0. ); 
//	};
	
	/** standardized view of regularly-spaced wavenumbers */
	struct wnums_t
	{
	    wnum_t baseWnum;
	    wnum_t deltaWnum;
	    unsigned wnumCount;
	    
	    /** truncate to bounds. FIXME: this should raise an exception if at all possible */
	    inline unsigned trunc( unsigned offset ) const 
		{ return (offset>=wnumCount)? wnumCount-1 : offset; }
	    
	    /** access as if it were an array */
	    inline double operator[]( unsigned offset ) const
		{ return deltaWnum*double(trunc(offset))+baseWnum ; }
	};

	bool operator== (const wnums_t&, const wnums_t&);
	
// FIXME - send this to the great library in the sky...		
	/** \brief Transforms a Wnums_t class into a blackbody radiance spectrum
	 * Future: this can create a cache full of NANs internally, and return values from the cache
	 * instead of recomputing the power every time an access occurs. 
	 * Possibly as an alternate implementation of the same interface, e.g. CachingPlanckRadianceGenerator
	 */
//	class PlanckRadianceGenerator
//	{
//	    const wnums_t &W;
//	    const kelvin_t T;
//	    
//	public:
//	    PlanckRadianceGenerator( const wnums_t& _w, const kelvin_t& _t ): 
//		W(_w), T(_t)
//	    { };
//		
//		inline radiance_t operator[]( unsigned offset ) const
//		{
//		    return planckRadiance( T, W[offset] ); 
//		};
//	};
//	
    } // namespace physics
} // namespace gips

#endif

