/** 
*  \file Connector.h
*  \brief landing point for connector-related definitions
*  \author R.K.Garcia <rayg@ssec.wisc.edu>
*
*  \version $Id: Connector.h,v 1.31 2005/07/08 07:12:52 maciek Exp $
*
*
*  Copyright UW/SSEC, ALL RIGHTS RESERVED, 2004
*  Space Science and Engineering Center
*  University of Wisconsin - Madison, USA
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this library; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
*  \endverbatim
*/

/**
* \mainpage UML-RT connectors implemented the C++ Template-fu way
 *
 * \par SimpleConnector diagram
 * The following diagram shows how the simple connector (whose payload, input and output
 * data structures are one and the same, set at compile time) is built up from its component
 * templated interfaces.
 * \image html simpleconnector.jpg
 * */


#ifndef H_CONNECTOR
#define H_CONNECTOR

#ifndef NO_DEPENDENCY_INCLUDES
#include "ProtocolRole.h"
#endif

///  \brief Connector namespace, including basic policies
namespace Connector 
{ 
   

    /**
     * Simple Connector wraps a connector transparently around any existing struct or object.
     * It is valid for as long as the wrapped structure is initialized
     */
    template <typename T>
    class SimpleConnector: public InputPort<T>, public OutputPort<T> 
    {
	T& payload;
    public:
	SimpleConnector(T& p): payload(p){}
	const T& getForInput() { return payload; }
	T& getForOutput() { return payload; }
    }; // class SimpleConnector


    template <typename P, typename C>
    class InputPortAdapter: public InputPort<C>
    {
        InputPort<P>* pcPtr;
    public:
        explicit InputPortAdapter(InputPort<P>& p): pcPtr(&p) {}
        //const C& getForInput() { return (C)(pcPtr->getForInput()); }
        const C& getForInput() { return *((const C *)&(pcPtr->getForInput())); }
    }; // class InputPortAdapter
        
    template <typename P, typename C>
    class OutputPortAdapter: public OutputPort<C>
    {
        OutputPort<P>* pcPtr;
    public:
        explicit OutputPortAdapter(OutputPort<P>& p): pcPtr(&p) {}
        C& getForOutput() { return *((C *)&(pcPtr->getForOutput())); }
    }; // class OutputPortAdapter
        

    
    /**
     * BlitzClosedConnector provides an internally-managed blitz array-backed connector.
     * Note that the constructor takes the size and makes its own private blitz array.
     */
    template <typename BT>
    class BlitzClosedConnector: public BlitzInputPort<BT>, public BlitzOutputPort<BT>
    {
        blitz::Array< BT, 1> backingVector;
    public:
        explicit BlitzClosedConnector( int size ): backingVector( size ) {};
	    virtual ~BlitzClosedConnector() {};
	    
	    blitz::Array<BT, 1>& getForOutput() { return backingVector; }
	    const blitz::Array<BT, 1>& getForInput() { return backingVector; }
    }; // class BlitzSimpleConnector
    

    
    /**
     * BlitzOpenConnector is like the BlitzSimpleConnector, except that its backing blitz array is supplied 
     * by the external initializer.
     */
    template <typename BT>
    class BlitzOpenConnector: public BlitzInputPort<BT>, public BlitzOutputPort<BT>
    {
        blitz::Array<BT, 1>& backingVector;
    public:
        explicit BlitzOpenConnector(blitz::Array<BT, 1>& outsider): backingVector(outsider) {};
	    virtual ~BlitzOpenConnector() {};
	    
	    blitz::Array<BT, 1>& getForOutput() { return backingVector; }
	    const blitz::Array<BT, 1>& getForInput() { return backingVector; }
    }; // class BlitzSimpleConnector



    /** a port wired to always send the same data - useful for initialization settings,
        guaranteed not to change in the course of runtime */
    template <class T>
    class ConstantInputPort: public InputPort<T>
    {
        T payload_;
    public:
        ConstantInputPort(const T& p): payload_(p) {}
	virtual ~ConstantInputPort() {};
	const T& getForInput() { return payload_; }; 
    }; // class ConstantInputPort
    
}; // namespace

#endif // ifndef H_CONNECTOR









