/**
 * \file Cifg2SpectrumMmapAdapter.cc
 * \brief Mmap adapter for Cifg2Spectrum.
 *
 * \author Maciek Smuga-Otto <maciek@ssec.wisc.edu>
 *
 * C++ end of UW-SSEC pipeline runtime environment. vector data are assumed passed in and out
 * via memory maps, while metadata are passed in and out via stdin/stdout.
 *
 *  \version $Id: Cifg2SpectrumMmapAdapter_GIFTSv1.cc,v 1.3 2005/07/08 20:21:48 maciek Exp $
 *
 *  \par Copyright:
 *  \verbatim
 *
 *  Copyright UW/SSEC, ALL RIGHTS RESERVED, 2004
 *  Space Science and Engineering Center
 *  University of Wisconsin - Madison, USA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 *  \endverbatim
 *
 */

#include <cstdlib>

#include <iostream>
#include <map>
#include <string>
#include <sstream>

#include "boost/shared_ptr.hpp"

#include "Physics.h"
#include "Instrument.h"
#include "Connector.h"
#include "MmapWrap.h"
#include "Cifg2Spectrum_GIFTSv1.h"

using namespace std;
using namespace gips;

// this is the adapter to the python surroundings, via memory maps and stdin/stdout
class PipelineEnv
{
    const string instrStr, bandStr;     // instrument string, band string.     
    const int bandSize;
    
    MmapWrap<Cifg2Spectrum::settings_t> settingsMmap;
    MmapWrap<Cifg2Spectrum::obs_in_t> obs_inMmap;
    MmapWrap<Cifg2Spectrum::obs_out_t> obs_outMmap;
    
    Connector::SimpleConnector<Cifg2Spectrum::settings_t> settings;
    Connector::SimpleConnector<Cifg2Spectrum::obs_in_t> obs_in;
    Connector::SimpleConnector<Cifg2Spectrum::obs_out_t> obs_out;

    Cifg2Spectrum::Ports ports;

    typedef boost::shared_ptr<Cifg2Spectrum> PipelinePtr;
    map<string, PipelinePtr> pipelines;

    
public:

    PipelineEnv():
	settingsMmap("GIPS_SETTINGS_MMAP"),
	obs_inMmap("GIPS_DATA_IN_MMAP"),
        obs_outMmap("GIPS_DATA_OUT_MMAP", true),

        settings(*(settingsMmap.getRawPtr())),
        obs_in(*(obs_inMmap.getRawPtr())),
        obs_out(*(obs_outMmap.getRawPtr())),

	ports(settings, obs_in, obs_out) {}

    ~PipelineEnv() {}

    void createPipeline(const string& pipelineKey)
	{
	    // cerr << "createPipeline: " << (ports.settings.getForInput()).npts << endl; // DEBUG

	    pipelines[pipelineKey] = PipelinePtr(new Ci2sImpl(ports));
	    cout << "CREATED:" << pipelineKey << endl;
	} 


    void runPipeline(const string& pipelineKey)
	{
	    PipelinePtr pipeline(pipelines[pipelineKey]);
	    
	    //cerr << "C++: ready to cycle pipeline operator" << endl; // DEBUG
	    //cerr << "C++: in@600 " << (obs_inMmap.getRawPtr()->rawInterferogram)[600] << endl; // DEBUG

	    Operator::Result res = (*pipeline)();
	    assert(res == Operator::OK);	

	    //cerr << "C++: out@600 " << (obs_outMmap.getRawPtr()->rawSpectrum)[600] << endl; // DEBUG
	    
	    cout << "PROCESSED:" <<  pipelineKey << endl; 
	}
    
};


int main()
{
    //cerr << "C++ alive and kicking" << endl; // DEBUG
    // construct PipelineEnv from supplied memory maps
    PipelineEnv penv;
    string commandStr;

    while(getline(cin, commandStr))
    {    
        //cerr << "C++ pipeline got:" << commandStr << endl; // DEBUG
	istringstream ssin(commandStr);
	string commandName;
	ssin >> commandName;
	  if (commandName == "CREATE")
	{
	  //cerr << "creating pipeline" << endl;
	    // creating pipeline
	    penv.createPipeline("IFG"); // HACK?
	}
	  else if (commandName == "PROCESS")
	{	    
	  //cerr << "running pipeline" << endl;
	    // pipeline setup
	    penv.runPipeline("IFG");
	}
	else
	    assert("currently only CREATE and PROCESS commands are accepted");
    }
    return(0);
};
