/*
 * \file CplxFFT.cc
 * \brief SHIS implementation of CplxFFT operator
 * \author Maciek Smuga-Otto <maciek@ssec.wisc.edu>
 *
 * based directly on the Scanning-HIS rsh2fbf application, written by Ray Garcia <rayg@ssec.wisc.edu>
 *
 *  \version $Id: CplxFFT.cc,v 1.26 2005/07/08 07:12:52 maciek Exp $
 *
 *  \par Copyright:
 *  \verbatim
 *
 *  Copyright UW/SSEC, ALL RIGHTS RESERVED, 2004
 *  Space Science and Engineering Center
 *  University of Wisconsin - Madison, USA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 *  \endverbatim
 */


#include "CplxFFT.h"
#include <iostream>

namespace gips {

CplxFFT::CplxFFT(Ports &p): ports(p)
{
    // FIXME: is getForInput() semantically correct here, considering as its values are only 
    // guaranteed constant over the course of a single operator() invocation.
    const settings_t &settings(ports.settings.getForInput());
    assert( settings.shiftedInterferogramSize == settings.doubledRawSpectrumSize );

    plan = fftw_create_plan( settings.shiftedInterferogramSize, FFTW_FORWARD, FFTW_MEASURE ); // alt: FFTW_ESTIMATE
    fftw_threads_init();    
}; // ctor

CplxFFT::~CplxFFT() {
    fftw_destroy_plan( plan );
}; // dtor


Operator::Result CplxFFT::operator() ( )
{
    const settings_t &settings(ports.settings.getForInput());
    const obs_in_t &obs_in(ports.obs_in.getForInput());
    obs_out_t & obs_out(ports.obs_out.getForOutput());
    
    // sanity check on buffers
    assert( sizeof(obs_in.shiftedInterferogram) 
	    >= sizeof(complex<double>) * settings.shiftedInterferogramSize ); 
    assert( sizeof(obs_out.doubledRawSpectrum) 
	    >= sizeof(complex<double>) * settings.doubledRawSpectrumSize ); 

    fftw_complex *interferogram_in = (fftw_complex *)(obs_in.shiftedInterferogram);
    fftw_complex *spectrum_out = (fftw_complex *)(obs_out.doubledRawSpectrum);

    //std::cerr << "ifg in length: " << si.size() << " spec out length: " << pfs.size() << "\n";
    // std::cerr << "FFT input channels 0,10,200: " << si(0) << "," << si(10) << "," << si(200) << "\n";
    // not using the threaded version due to compilation errors
    fftw_one( plan, interferogram_in, spectrum_out );
    // std::cerr << "FFT output channels 0,10,200: " << pfs(0) << "," << pfs(10) << "," << pfs(200) << "\n";

    //std::cerr << "CplxFFT: done" << std::endl;

    return Operator::OK;
};

} // namespace gips

