/**
 * \file IfgShift.cc
 * \brief SHIS implementation of IfgShift operator
 * \author Maciek Smuga-Otto <maciek@ssec.wisc.edu>
 *
 * based directly on the Scanning-HIS rsh2fbf application, written by Ray Garcia <rayg@ssec.wisc.edu>
 *
 *  \version $Id: IfgShift.cc,v 1.25 2005/07/08 07:12:52 maciek Exp $
 *
 *  \par Copyright:
 *  \verbatim
 *
 *  Copyright UW/SSEC, ALL RIGHTS RESERVED, 2004
 *  Space Science and Engineering Center
 *  University of Wisconsin - Madison, USA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 *  \endverbatim
 */
#include "IfgShift.h"

#include <iostream>

using namespace gips;

IfgShift::IfgShift(IfgShift::Ports &ports_): ports(ports_)
{
    const settings_t& settings(ports.settings.getForInput());
    assert( settings.rawInterferogramSize == settings.shiftedInterferogramSize );
}

IfgShift::~IfgShift() {}

/// \brief Concrete interferogram shift operator
/// 
/// transforms vector v_0 ... v_m-1 v_m ... v_npts 
/// into vector v_m ... v_npts v0 ... v_m-1
/// where m = npts/2
Operator::Result IfgShift::operator()( )
{
    //std::cerr << "C++: in IfgShiftOperator::operator()" << std::endl;

    // Block of interface code, getting access to ports.
    const settings_t& settings(ports.settings.getForInput());
    const obs_in_t& obs_in(ports.obs_in.getForInput());
    obs_out_t& obs_out(ports.obs_out.getForOutput());

    const instrument::BlitzCplxDoubleVector in(const_cast<complex<double>*>(&(obs_in.rawInterferogram[0])), 
                                               blitz::shape(settings.rawInterferogramSize),
                                               blitz::neverDeleteData);
    instrument::BlitzCplxDoubleVector out(&(obs_out.shiftedInterferogram[0]),
                                          blitz::shape(settings.shiftedInterferogramSize),
                                          blitz::neverDeleteData);

    size_t N(settings.rawInterferogramSize), halfN(N/2); // we're using std::cplx, not interleaving

    //std::cerr << "IfgShift: record #" << ancillary.recordNum << std::endl;
    //std::cerr << "C++: about to perform computation with N=" << N << std::endl;
        
    if ( !N || N%2 ) return Operator::ERROR;

    //std::cerr << "IfgShift: N=" << settings.npts << std::endl;

    unsigned src(0), dest(0);
    // again, dealing this time with std::cplx members, not interleaved real/imag parts
    for( src = halfN; src<N; src++,dest++ ) {
        out(dest) = in(src);
    }
    for( src=0; src<halfN; src++, dest++ ) {
        out(dest) = in(src); 
    }    

    return Operator::OK;
}


/******************** original function body in rsh2fbf:
int sol1IfgShift( solContent &mcontent, const solContent &C ) {
    size_t N = C.size()/2, halfN = N/2;    
    if (!N || N%2) return -1;
    mcontent.resize( C.size() ); // solContent inherits from std::vector
    size_t src, dest=0;
    // fftshift( data ) for complex data
    for( src = halfN; src<N; src++,dest++ ) {
        mcontent[dest*2] = C[src*2];
        mcontent[(dest*2)+1] = C[(src*2)+1];
    }
    for( src=0; src<halfN; src++, dest++ ) {
        mcontent[dest*2] = C[src*2]; 
        mcontent[(dest*2)+1] = C[(src*2)+1];
    }
    return 0;
}
*********************/

